import pathlib
import sys

sys.path.append(str(pathlib.Path(__file__).parent))

extensions = [
    "sphinx.ext.autodoc",
    "sphinx_rtd_theme",
    "sphinx_toolbox.more_autodoc.autoprotocol",
]

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_css_files = ["css/extra.css"]

exclude_patterns = ["_build/**", ".sphinx-build/**", "README.rst"]

master_doc = "index"
source_suffix = ".rst"

pygments_style = "pastie"

copyright = "Stephen Moore"
project = "extended_mypy_django_plugin"

version = "0.1"
release = "0.1"

autodoc_preserve_defaults = True



"""
This exists as a helper for defining the mypy plugin.

The way a mypy plugin works is there is a class that inherits from
``mypy.plugin.Plugin`` with specific hooks that take in a string and returns
a function.

So for example, the ``get_customize_class_mro_hook`` hook will take in the
``fullname`` representing classes that can be altered, and a function that takes
in a ``ClassDefContext`` object and returns ``None`` must be returned if
the plugin wants to do something with that object.

The first plugin that mypy encounters which returns a function will win and no
other plugins will get to look at that fullname.

This is fine but it can get awkward when the function returned takes in more
options than only the context. To improve this situation, this module implements
a ``Hook`` class and associated decorator to turn those hooks into python
descriptors that do the correct thing.

.. code-block:: python

    from typing import Generic
    from mypy.plugin import Plugin, AttributeContext
    from mypy.types import Type as MypyType
    from extended_mypy_django_plugin.plugin import _hook



    class Hook(
        Generic[_hook.T_Ctx, _hook.T_Ret], _hook.Hook["MyPlugin", _hook.T_Ctx, _hook.T_Ret]
    ): ...


    class MyPlugin(Plugin):
        @_hook.hook
        class get_attribute_hook(Hook[AttributeContext, MypyType]):
            def choose(self) -> bool:
                # return True if we want to use the run method for ``self.fullname``.
                return self.fullname.endswith(".blah")

            def run(self, ctx: AttributeContext) -> MypyType:
                # Do stuff
                return ...

The hook has ``choose`` which is essentially what would be on the plugin itself
in normal situations. It can find the string that would be passed in as
``self.fullname``. It returns ``True`` if we want to handle this string, otherwise
``False``.

If the hook is chosen, then ``run`` is ran with the specified context class and
needs to return what is needed from it.

The hook has access to it:

plugin
    The plugin instance itself.

fullname
    The string that was passed into the plugin's hook

super_hook
    What would have been returned by the class the plugin is inheriting from.
"""

from __future__ import annotations

import abc
from collections.abc import Callable
from typing import Generic, TypeAlias, TypeVar, overload

from mypy.plugin import Plugin

T_Ctx = TypeVar("T_Ctx")
T_Ret = TypeVar("T_Ret")
T_Plugin = TypeVar("T_Plugin", bound=Plugin)


_HookChooser: TypeAlias = Callable[[str], Callable[[T_Ctx], T_Ret] | None]


class Hook(Generic[T_Plugin, T_Ctx, T_Ret], abc.ABC):
    """
    Class used to represent both the choosing and running logic for a hook
    on a mypy Plugin.

    This is to be used with the ``hook`` descriptor defined below. See the
    docstring on the module for more information.

    Concrete subclasses of this must implement ``choose`` and ``run``.

    The ``choose`` method will be used to determine if this hook should be used
    by mypy to do something.

    The ``run`` method will take in the context provided by mypy if this hook
    was chosen. The type of that context and what needs to be returned is defined
    by the specific mypy hook being implemented, and should be specified as
    type vars when defining the implementation of the hook.
    """

    def __init__(
        self,
        plugin: T_Plugin,
        fullname: str,
        super_hook: Callable[[T_Ctx], T_Ret] | None,
    ) -> None:
        self.plugin = plugin
        self.fullname = fullname
        self.super_hook = super_hook
        self.extra_init()

    def extra_init(self) -> None: ...

    @abc.abstractmethod
    def run(self, ctx: T_Ctx) -> T_Ret: ...

    @abc.abstractmethod
    def choose(self) -> bool: ...

    def hook(self) -> Callable[[T_Ctx], T_Ret] | None:
        """
        This is the function that mypy ends up calling when asking the plugin
        if it should handle something.
        """
        if self.choose():
            return self.run
        else:
            return self.super_hook


class hook(Generic[T_Plugin, T_Ctx, T_Ret]):
    """
    This is a descriptor classes used to return a callable object that takes in
    a string and either returns a function that mypy can use to perform an action,
    or return None if this hook does not need to do anything in that instance.
    """

    def __init__(self, hook: type[Hook[T_Plugin, T_Ctx, T_Ret]]) -> None:
        self.hook = hook
        self.__doc__ = hook.__doc__

    name: str

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name
        self.owner = owner

    @overload
    def __get__(self, instance: None, owner: None) -> hook[T_Plugin, T_Ctx, T_Ret]: ...

    @overload
    def __get__(self, instance: T_Plugin, owner: type[T_Plugin]) -> _HookChooser[T_Ctx, T_Ret]: ...

    def __get__(
        self, instance: T_Plugin | None, owner: type[T_Plugin] | None = None
    ) -> hook[T_Plugin, T_Ctx, T_Ret] | _HookChooser[T_Ctx, T_Ret]:
        if instance is None:
            return self

        super_hook = getattr(super(self.owner, instance), self.name)

        def result(fullname: str) -> Callable[[T_Ctx], T_Ret] | None:
            return self.hook(
                plugin=instance,
                fullname=fullname,
                super_hook=super_hook(fullname),
            ).hook()

        return result



import inspect
import os
import shlex
import shutil
import sys
import typing as tp
from collections.abc import Callable
from pathlib import Path

here = Path(__file__).parent


class Command:
    __is_command__: bool

    def __call__(self, bin_dir: Path, args: list[str]) -> None: ...


def command(func: Callable[..., None]) -> Callable[..., None]:
    tp.cast(Command, func).__is_command__ = True
    return func


def run(*args: str | Path) -> None:
    cmd = " ".join(shlex.quote(str(part)) for part in args)
    print(f"Running '{cmd}'\n")  # noqa: T201
    ret = os.system(cmd)
    if ret != 0:
        sys.exit(1)


class App:
    commands: dict[str, Command]

    def __init__(self) -> None:
        self.commands = {}

        compare = inspect.signature(type("C", (Command,), {})().__call__)

        for name in dir(self):
            val = getattr(self, name)
            if getattr(val, "__is_command__", False):
                assert (
                    inspect.signature(val) == compare
                ), f"Expected '{name}' to have correct signature, have {inspect.signature(val)} instead of {compare}"
                self.commands[name] = val

    def __call__(self, args: list[str]) -> None:
        bin_dir = Path(sys.executable).parent

        if args and args[0] in self.commands:
            os.chdir(here.parent)
            self.commands[args[0]](bin_dir, args[1:])
            return

        sys.exit(f"Unknown command:\nAvailable: {sorted(self.commands)}\nWanted: {args}")

    @command
    def lint(self, bin_dir: Path, args: list[str]) -> None:
        run(bin_dir / "ruff", "check", *args)

    @command
    def format(self, bin_dir: Path, args: list[str]) -> None:
        if not args:
            args = [".", *args]
        run(bin_dir / "ruff", "format", *args)
        run(bin_dir / "ruff", "check", "--fix", "--select", "I,UP", *args)

    @command
    def types(self, bin_dir: Path, args: list[str]) -> None:
        specified: bool = True
        locations: list[str] = [a for a in args if not a.startswith("-")]
        args = [a for a in args if a.startswith("-")]

        if not locations:
            specified = False
            locations.append(str((here / "..").resolve()))
        else:
            cwd = Path.cwd()
            paths: list[Path] = []
            for location in locations:
                from_current = cwd / location
                from_root = here.parent / location

                if from_current.exists():
                    paths.append(from_current)
                elif from_root.exists():
                    paths.append(from_root)
                else:
                    raise ValueError(f"Couldn't find path for {location}")

            example_root = here.parent / "example"
            if any(path.is_relative_to(example_root) for path in paths):
                if not all(path.is_relative_to(example_root) for path in paths):
                    raise ValueError("If specifying an example path, all paths must be from there")
                os.chdir(example_root)
            locations = [str(path) for path in paths]

        run(bin_dir / "mypy", *locations, *args)

        if not specified:
            os.chdir(here.parent / "example")
            run(bin_dir / "mypy", ".", *args)

    @command
    def tests(self, bin_dir: Path, args: list[str]) -> None:
        run(bin_dir / "pytest", *args)

    @command
    def docs(self, bin_dir: Path, args: list[str]) -> None:
        docs_path = here / ".." / "docs"
        build_path = docs_path / "_build"
        command: list[Path | str] = [bin_dir / "sphinx-build"]

        other_args: list[str] = []
        for arg in args:
            if arg == "fresh":
                if build_path.exists():
                    shutil.rmtree(build_path)
            elif arg == "view":
                command = [bin_dir / "sphinx-autobuild", "--port", "9876"]
            else:
                other_args.append(arg)

        os.chdir(docs_path)

        run(*command, ".", "_build/html", "-b", "html", "-d", "_build/doctrees", *other_args)


app = App()

if __name__ == "__main__":
    app(sys.argv[1:])



from mypy.nodes import TypeInfo
from mypy.semanal import SemanticAnalyzer
from mypy.typeanal import TypeAnalyser
from mypy.types import (
    AnyType,
    Instance,
    TypeOfAny,
    TypeVarType,
    UnboundType,
    UnionType,
    get_proper_type,
)
from mypy.types import (
    Type as MypyType,
)

from .. import _store


class TypeAnalyzing:
    def __init__(
        self, store: _store.Store, *, api: TypeAnalyser, sem_api: SemanticAnalyzer
    ) -> None:
        self.api = api
        self.sem_api = sem_api
        self.store = store

    def find_concrete_models(self, unbound_type: UnboundType) -> MypyType:
        args = unbound_type.args
        type_arg = get_proper_type(self.api.analyze_type(args[0]))

        if not isinstance(type_arg, Instance):
            return UnionType(())

        concrete = tuple(
            self.store.retrieve_concrete_children_types(
                type_arg.type, self.lookup_info, self.sem_api.named_type_or_none
            )
        )
        if not concrete:
            if self.sem_api.final_iteration:
                self.api.fail(
                    f"No concrete models found for {type_arg.type.fullname}", unbound_type
                )
                return AnyType(TypeOfAny.from_error)
            else:
                self.sem_api.defer()
                return unbound_type

        return UnionType(concrete)

    def find_concrete_querysets(self, unbound_type: UnboundType) -> MypyType:
        args = unbound_type.args
        type_arg = get_proper_type(self.api.analyze_type(args[0]))

        if not isinstance(type_arg, Instance):
            return UnionType(())

        concrete = tuple(
            self.store.retrieve_concrete_children_types(
                type_arg.type, self.lookup_info, self.sem_api.named_type_or_none
            )
        )
        if not concrete:
            self.api.fail(f"No concrete models found for {type_arg.type.fullname}", unbound_type)
            return AnyType(TypeOfAny.from_error)

        try:
            querysets = tuple(self.store.realise_querysets(UnionType(concrete), self.lookup_info))
        except _store.RestartDmypy:
            self.api.fail("You probably need to restart dmypy", unbound_type)
            return AnyType(TypeOfAny.from_error)
        except _store.UnionMustBeOfTypes:
            self.api.fail("Union must be of instances of models", unbound_type)
            return AnyType(TypeOfAny.from_error)
        else:
            return UnionType(querysets)

    def find_default_queryset(self, unbound_type: UnboundType) -> MypyType:
        args = unbound_type.args
        type_arg = get_proper_type(self.api.analyze_type(args[0]))

        if isinstance(type_arg, AnyType):
            self.api.fail("Can't get default query set for Any", unbound_type)
            return unbound_type

        if isinstance(type_arg, TypeVarType):
            return unbound_type

        if not isinstance(type_arg, Instance | UnionType):
            self.api.fail("Default queryset needs a class to find for", unbound_type)
            return unbound_type

        try:
            querysets = tuple(self.store.realise_querysets(type_arg, self.lookup_info))
        except _store.RestartDmypy:
            self.api.fail("You probably need to restart dmypy", unbound_type)
            return AnyType(TypeOfAny.from_error)
        except _store.UnionMustBeOfTypes:
            self.api.fail("Union must be of instances of models", unbound_type)
            return unbound_type
        else:
            return UnionType(querysets)

    def lookup_info(self, fullname: str) -> TypeInfo | None:
        instance = self.sem_api.named_type_or_none(fullname)
        if instance:
            return instance.type

        return self.store._plugin_lookup_info(fullname)



import sys


def debug(*args: object) -> None:
    """
    Helper to print debug information to sys.stderr in a format
    that the pytest-mypy-plugins pytest plugin will ignore
    """
    print(":debug:", *args, file=sys.stdout)



from collections.abc import MutableMapping

from mypy.options import Options
from mypy.plugin import Plugin as MypyPlugin

from .plugin import ExtendedMypyStubs


class PluginProvider:
    """
    This can be used to provide both a mypy plugin as well as a __version__ that changes
    when mypy needs to do a full restart.

    Given either the extended_mypy_django_plugin.plugin.ExtendedMypyStubs class or a subclass
    of that, usage is::

        from extended_mypy_django_plugin.plugin import ExtendedMypyStubs
        from extended_mypy_django_plugin.entry import PluginProvider

        plugin = PluginProvider(ExtendedMypyStubs, locals())
    """

    def __init__(
        self, plugin_cls: type[ExtendedMypyStubs], locals: MutableMapping[str, object], /
    ) -> None:
        self.locals = locals
        self.instance: ExtendedMypyStubs | None = None
        self.plugin_cls = plugin_cls
        self.previous_version: int | None = None

    def _change_version(self, instance: ExtendedMypyStubs) -> None:
        new_version = instance.determine_plugin_version(self.previous_version)
        self.previous_version = new_version
        self.locals["__version__"] = str(new_version)

    def __call__(self, version: str) -> type[MypyPlugin]:
        if self.instance is not None:
            self._change_version(self.instance)

            # Inside dmypy, don't create a new plugin
            return MypyPlugin

        provider = self
        major, minor, _ = version.split(".", 2)

        def __init__(instance: ExtendedMypyStubs, options: Options) -> None:
            super(instance.__class__, instance).__init__(
                options,
                mypy_version_tuple=(int(major), int(minor)),  # type: ignore[call-arg]
            )
            provider.instance = instance
            provider._change_version(instance)

        return type("Plugin", (provider.plugin_cls,), {"__init__": __init__})



# It is used in `mypy.ini` only.
INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.admin",
    "django.contrib.contenttypes",
    "django.contrib.flatpages",
    "django.contrib.redirects",
    "django.contrib.sessions",
    "django.contrib.sites",
]



import pathlib
from collections.abc import Iterator
from typing import TYPE_CHECKING

from django.db import models

if TYPE_CHECKING:
    from django.apps.registry import Apps


def record_known_models(output_location: pathlib.Path, apps_registry: "Apps") -> None:
    found: set[str] = set()
    for model in find_known_models(apps_registry):
        found.add(f"{model.__module__}.{model.__qualname__}")

    output_location.write_text("\n".join(sorted(found)))


def find_known_models(apps_registry: "Apps") -> Iterator[type[models.Model]]:
    for concrete_model_cls in apps_registry.get_models():
        yield concrete_model_cls

        # And find abstract classes
        for model_cls in concrete_model_cls.mro()[1:]:
            if (
                issubclass(model_cls, models.Model)
                and hasattr(model_cls, "_meta")
                and model_cls._meta.abstract
            ):
                yield model_cls



import ast
import dataclasses
import os
import pathlib
import runpy
from collections.abc import Mapping, MutableSequence
from typing import TYPE_CHECKING

from pytest_mypy_plugins import (
    ExtensionHook,
    FollowupFile,
    ItemForHook,
    MypyPluginsScenario,
    OutputMatcher,
    ScenarioHookMaker,
    ScenarioHooks,
    ScenarioHooksRunAndCheckOptions,
)

here = pathlib.Path(__file__).parent


def django_plugin_hook(item: ItemForHook) -> None:
    django_settings_section = (
        "\n[mypy.plugins.django-stubs]\n" "django_settings_module = mysettings"
    )

    if not item.additional_mypy_config:
        item.additional_mypy_config = django_settings_section
    else:
        if "[mypy.plugins.django-stubs]" not in item.additional_mypy_config:
            item.additional_mypy_config += django_settings_section


class Hooks(ScenarioHooks):
    def before_run_and_check_mypy(
        self,
        *,
        scenario: MypyPluginsScenario,
        options: ScenarioHooksRunAndCheckOptions,
        config_file: pathlib.Path,
        expected_output: MutableSequence[OutputMatcher],
        additional_properties: Mapping[str, object],
    ) -> ScenarioHooksRunAndCheckOptions:
        custom_settings = additional_properties.get("custom_settings", None)
        copied_apps = additional_properties.get("copied_apps", None)
        installed_apps = additional_properties.get("installed_apps", None)
        monkeypatch = additional_properties.get("monkeypatch", None)

        options = dataclasses.replace(options, start=["."])

        if "debug" in additional_properties:
            pathlib.Path("/tmp/debug").write_text("")
        else:
            pathlib.Path("/tmp/debug").unlink(missing_ok=True)

        if isinstance(copied_apps, list):
            for app in copied_apps:
                if isinstance(app, str):
                    self._copy_app(scenario, app)

        current_settings: pathlib.Path = scenario.path_for("mysettings.py")

        if custom_settings is None and installed_apps is None and monkeypatch is None:
            if current_settings.exists():
                return options

        if not current_settings.exists():
            with open(current_settings, "w") as fle:
                fle.write("")

        original_settings_text = current_settings.read_text()

        if monkeypatch is not None:
            monkeypatch_str = "import django_stubs_ext\ndjango_stubs_ext.monkeypatch()\n"
            settings_text = original_settings_text.replace(monkeypatch_str, "")
            if monkeypatch:
                settings_text = monkeypatch_str + settings_text

            with open(current_settings, "w") as fle:
                fle.write(settings_text)

        settings_values = runpy.run_path(str(current_settings))

        if "INSTALLED_APPS" not in settings_values:
            with open(current_settings, "a") as fle:
                fle.write("\nINSTALLED_APPS = None")

        if "SECRET_KEY" not in settings_values:
            with open(current_settings, "a") as fle:
                fle.write('\nSECRET_KEY = "1"')

        settings = ast.parse(current_settings.read_text())

        class Fixer(ast.NodeTransformer):
            def visit_Assign(self, node: ast.Assign) -> ast.Assign:
                match node.targets:
                    case [ast.Name(id="INSTALLED_APPS")]:
                        nonlocal installed_apps
                        if not isinstance(node.value, ast.List):
                            installed_apps = installed_apps or ["myapp", "myapp2"]
                        elif installed_apps is None and isinstance(
                            settings_values["INSTALLED_APPS"], list
                        ):
                            installed_apps = settings_values["INSTALLED_APPS"]

                        if not isinstance(installed_apps, list):
                            installed_apps = []

                        if "django.contrib.contenttypes" not in installed_apps:
                            installed_apps.insert(0, "django.contrib.contenttypes")

                        return ast.Assign(
                            targets=node.targets,
                            value=ast.List(
                                elts=[ast.Constant(value=app) for app in installed_apps]
                            ),
                        )
                    case _:
                        return node

        Fixer().visit(settings)
        with open(current_settings, "w") as fle:
            fle.write(ast.unparse(ast.fix_missing_locations(settings)))
            if isinstance(custom_settings, str):
                fle.write("\n")
                fle.write(custom_settings)

        if current_settings.read_text() != original_settings_text:
            new_settings = current_settings.read_text()
            current_settings.write_text("")
            # Make it recorded that the settings changed
            scenario.handle_followup_file(
                FollowupFile(path=current_settings.name, content=new_settings)
            )

        settings_values = runpy.run_path(str(current_settings))
        installed_apps = settings_values["INSTALLED_APPS"]

        if not isinstance(installed_apps, list):
            installed_apps = []

        for app in installed_apps:
            if (here / app).exists():
                self._copy_app(scenario, app)

        return options

    def _copy_app(self, scenario: MypyPluginsScenario, app: str) -> None:
        for root, _, files in os.walk(here / app):
            for name in files:
                location = pathlib.Path(root, name)
                path = location.relative_to(here)
                if not (pathlib.Path.cwd() / path).exists():
                    scenario.handle_followup_file(
                        FollowupFile(path=str(path), content=location.read_text())
                    )


if TYPE_CHECKING:
    _eh: ExtensionHook = django_plugin_hook
    _sh: ScenarioHookMaker = Hooks



#!/usr/bin/env python

import argparse
import os
import pathlib
import sys
from collections.abc import Callable

from extended_mypy_django_plugin.scripts import record_known_models


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--django-settings-module", help="The setting value to use for DJANGO_SETTINGS_MODULE"
    )
    parser.add_argument(
        "--apps-file",
        help="The file to print the installed apps to, one app per line",
        type=pathlib.Path,
    )
    parser.add_argument(
        "--known-models-file",
        help="The file to print the known models to",
        type=pathlib.Path,
    )
    parser.add_argument(
        "--scratch-path",
        help="The folder that the plugin is allowed to write in",
        type=pathlib.Path,
    )
    return parser


def main(
    argv: list[str] | None = None, additional_django_setup: Callable[[], None] | None = None
) -> None:
    parser = make_parser()
    args = parser.parse_args(argv)

    if (args.scratch_path / "__assume_django_state_unchanged__").exists():
        sys.exit(2)

    os.environ["DJANGO_SETTINGS_MODULE"] = args.django_settings_module
    if additional_django_setup:
        additional_django_setup()

    # add current directory to sys.path
    sys.path.append(str(pathlib.Path.cwd()))

    from django.apps import apps
    from django.conf import settings

    if not settings.configured:
        settings._setup()  # type: ignore[misc]
    apps.populate(settings.INSTALLED_APPS)

    assert apps.apps_ready, "Apps are not ready"
    assert settings.configured, "Settings are not configured"

    args.apps_file.write_text("\n".join(settings.INSTALLED_APPS))
    record_known_models(args.known_models_file, apps)


if __name__ == "__main__":
    main()



from django.db import models


class Parent(models.Model):
    one = models.CharField(max_length=50)

    class Meta:
        abstract = True


class Child1(Parent):
    concrete_from_myapp = models.CharField(max_length=50)
    two = models.CharField(max_length=60)


class Child2QuerySet(models.QuerySet["Child2"]):
    pass


Child2Manager = models.Manager.from_queryset(Child2QuerySet)


class Child2(Parent):
    concrete_from_myapp = models.CharField(max_length=50)
    two = models.CharField(max_length=60)
    four = models.CharField(max_length=1)

    three = models.CharField(max_length=70)

    objects = Child2Manager()


class Parent2(Parent):
    three = models.CharField(max_length=51)

    class Meta:
        abstract = True


class Child3(Parent2):
    concrete_from_myapp = models.CharField(max_length=50)
    two = models.CharField(max_length=60)

    three = models.CharField(max_length=70)



import dataclasses
import importlib.resources
import importlib.util
import io
import itertools
import pathlib
import re
import shlex
import stat
import subprocess
import sys
import tempfile
import textwrap
import time
import zlib
from collections import defaultdict
from collections.abc import Iterator, Mapping, MutableMapping
from typing import Any, ClassVar, Protocol, Union

from django.db import models
from django.db.models.fields.related import ForeignObjectRel, RelatedField

ModelModules = Mapping[str, Mapping[str, type[models.Model]]]


regexes = {
    "mod_decl": re.compile(r'^mod = "(?P<mod>[^"]+)"$'),
    "summary_decl": re.compile(r'^summary = "(?P<summary>[^"]+)"$'),
}


class ModelRelatedFieldsGetter(Protocol):
    def __call__(self, model_cls: type[models.Model]) -> Iterator["RelatedField[Any, Any]"]: ...


class FieldRelatedModelClsGetter(Protocol):
    def __call__(
        self, field: Union["RelatedField[Any, Any]", ForeignObjectRel]
    ) -> type[models.Model]: ...


class ReportNamesGetter(Protocol):
    def __call__(self, fullname: str, deps: set[str], /) -> Iterator[str]: ...


@dataclasses.dataclass
class _Store:
    prefix: str
    reports_dir: pathlib.Path

    modules: Mapping[str, str] = dataclasses.field(default_factory=dict)
    modules_to_report_name: MutableMapping[str, str] = dataclasses.field(default_factory=dict)

    version: ClassVar[str] = "json.1"

    @classmethod
    def read(cls, prefix: str, reports_dir: pathlib.Path) -> "_Store":
        modules: dict[str, str] = {}
        if not reports_dir.exists():
            reports_dir.mkdir(parents=True, exist_ok=True)

        for path in reports_dir.iterdir():
            mod: str | None = None
            summary: str | None = None

            if path.suffix == ".py":
                for line in path.read_text().splitlines():
                    m = regexes["mod_decl"].match(line)
                    if m:
                        mod = m.groupdict()["mod"]

                    m = regexes["summary_decl"].match(line)
                    if m:
                        summary = m.groupdict()["summary"]

            if mod is None or summary is None:
                path.unlink()
            else:
                try:
                    spec = importlib.util.find_spec(mod)
                except ModuleNotFoundError:
                    spec = None

                if spec:
                    modules[mod] = summary
                else:
                    path.unlink()

        return cls(prefix=prefix, modules=modules, reports_dir=reports_dir).write(modules)

    def _write_mod(
        self, directory: pathlib.Path, mod: str, summary: str, empty: bool = False
    ) -> str:
        name = f"mod_{zlib.adler32(mod.encode())}"
        self.modules_to_report_name[mod] = f"{self.prefix}.{name}"

        # For mypy to trigger this dependency as stale it's interface must change
        # So we produce a different function each time using the current time
        content = textwrap.dedent(f"""
        def value_{'not_installed' if empty else str(time.time()).replace('.', '__')}() -> str:
            return ""

        mod = "{mod}"
        summary = "{summary}"
        """)

        destination = self.reports_dir / f"{name}.py"
        previous_summary: str | None = None
        if destination.exists():
            for line in destination.read_text().splitlines():
                m = regexes["summary_decl"].match(line)
                if m:
                    previous_summary = m.groupdict()["summary"]
                    break

        if summary != previous_summary:
            (directory / f"{name}.py").write_text(content)
        return name

    def add_mod(self, mod: str) -> str:
        with tempfile.TemporaryDirectory() as tmp:
            temp_dir = pathlib.Path(tmp)
            summary = f"{mod} ||>"
            name = self._write_mod(temp_dir, mod, summary, empty=True)
            made = temp_dir / f"{name}.py"
            if made.exists():
                made.rename(self.reports_dir / f"{name}.py")
            return f"{self.prefix}.{name}"

    def write(self, modules: Mapping[str, str]) -> "_Store":
        instance = self.__class__(
            prefix=self.prefix,
            modules=modules,
            reports_dir=self.reports_dir,
            modules_to_report_name=self.modules_to_report_name,
        )

        # Prevent partial writes by dumping to a temp directory and moving changed files
        # We also don't simply rename the entire directory so unchanged files remain unchanged
        with tempfile.TemporaryDirectory() as tmp:
            temp_dir = pathlib.Path(tmp)
            for mod, summary in instance.modules.items():
                instance._write_mod(temp_dir, mod, summary)

            for path in temp_dir.iterdir():
                destination = instance.reports_dir / path.name
                if not destination.exists():
                    path.rename(destination)
                elif path.read_bytes() != destination.read_bytes():
                    path.rename(destination)

        return instance


class _DepFinder:
    @classmethod
    def find_from(
        cls,
        model_modules: ModelModules,
        django_settings_module: str,
        get_model_related_fields: ModelRelatedFieldsGetter,
        get_field_related_model_cls: FieldRelatedModelClsGetter,
    ) -> tuple[Mapping[str, set[str]], Mapping[str, set[str]]]:
        instance = cls(model_modules=model_modules, django_settings_module=django_settings_module)
        found: set[str] = {django_settings_module}

        for mod, known in model_modules.items():
            found.add(mod)
            for name, model_cls in known.items():
                instance._find_models_in_mro(mod, model_cls)
                instance._find_related_models(
                    mod,
                    model_cls,
                    get_model_related_fields,
                    get_field_related_model_cls=get_field_related_model_cls,
                )

        result: dict[str, set[str]] = {}
        for mod in found:
            result[mod] = (
                set()
                .union(instance.known_models.get(mod) or set())
                .union(instance.related_models.get(mod) or set())
            )
        return result, instance.model_children

    def __init__(self, *, model_modules: ModelModules, django_settings_module: str) -> None:
        self.model_modules = model_modules
        self.django_settings_modules = django_settings_module

        self.related_models: dict[str, set[str]] = defaultdict(set)
        self.known_models: dict[str, set[str]] = defaultdict(set)
        self.model_children: dict[str, set[str]] = defaultdict(set)

    def _find_models_in_mro(self, mod: str, cls: type[models.Model]) -> None:
        cls_fullname = f"{cls.__module__}.{cls.__qualname__}"
        self.known_models[mod].add(cls_fullname)

        for mro in cls.mro():
            if mro is cls:
                continue
            if mro is models.Model:
                break

            mro_fullname = f"{mro.__module__}.{mro.__qualname__}"

            if mro.__module__ != mod:
                self.known_models[mod].add(mro_fullname)

            if not mro.__module__.startswith("django."):
                self.known_models[mro.__module__].add(cls_fullname)
                self.model_children[mro_fullname].add(cls_fullname)

    def _find_related_models(
        self,
        mod: str,
        cls: type[models.Model],
        get_model_related_fields: ModelRelatedFieldsGetter,
        get_field_related_model_cls: FieldRelatedModelClsGetter,
    ) -> None:
        cls_fullname = "{cls.__module__}.{cls.__qualname__}"

        for field in itertools.chain(
            # forward relations
            get_model_related_fields(cls),
            # reverse relations - `related_objects` is private API (according to docstring)
            cls._meta.related_objects,  # type: ignore[attr-defined]
        ):
            try:
                related_model_cls = get_field_related_model_cls(field)
            except Exception:
                continue

            else:
                if related_model_cls is None:
                    continue
            related_model_module = related_model_cls.__module__
            related_model_fullname = (
                f"{related_model_cls.__module__}.{related_model_cls.__qualname__}"
            )
            if related_model_module != mod:
                self.related_models[mod].add(related_model_fullname)
                if not related_model_module.startswith("django."):
                    self.related_models[related_model_module].add(cls_fullname)


class Reports:
    """
    This class is responsible for generating the files that represent our virtual
    dependencies.

    Further documentation has yet to be added.
    """

    @classmethod
    def create(
        cls,
        *,
        determine_django_state_script: pathlib.Path | None,
        django_settings_module: str,
        scratch_path: pathlib.Path,
        reports_dir_prefix: str = "__virtual_extended_mypy_django_plugin_report__",
    ) -> "Reports":
        if determine_django_state_script is not None:
            if not determine_django_state_script.exists():
                raise ValueError("The provided script for finding installed apps does not exist")

            if not determine_django_state_script.stat().st_mode & stat.S_IXUSR:
                raise ValueError(
                    "The provided script for finding installed apps is not executable!"
                )

        if determine_django_state_script is None:
            determine_django_state_script = pathlib.Path(
                str(
                    importlib.resources.files("extended_mypy_django_plugin")
                    / "scripts"
                    / "determine_django_state.py"
                )
            )

        reports_dir = scratch_path / reports_dir_prefix
        if reports_dir.exists() and not reports_dir.is_dir():
            reports_dir.unlink()
        reports_dir.mkdir(parents=True, exist_ok=True)

        return cls(
            store=_Store.read(prefix=reports_dir_prefix, reports_dir=reports_dir),
            determine_django_state_script=determine_django_state_script,
            django_settings_module=django_settings_module,
        )

    def __init__(
        self,
        *,
        store: _Store,
        determine_django_state_script: pathlib.Path,
        django_settings_module: str,
    ) -> None:
        self._store = store
        self._determine_django_state_script = determine_django_state_script
        self._django_settings_module = django_settings_module
        self._known_concrete_models: MutableMapping[str, set[str]] = defaultdict(set)

    def known_concrete_models(self, fullname: str) -> set[str]:
        return self._known_concrete_models[fullname]

    def lines_hash(self) -> str:
        buffer = io.BytesIO()
        for path in self._store.reports_dir.iterdir():
            content = path.read_bytes()
            if b"def value_not_installed" not in content:
                buffer.write(b"\n")
                buffer.write(path.name.encode())
                buffer.write(b"\n")
                buffer.write(path.read_bytes())

        return str(zlib.adler32(buffer.getbuffer()))

    def determine_version_hash(
        self, scratch_path: pathlib.Path, previous_version: int | None
    ) -> int:
        result_file_cm = tempfile.NamedTemporaryFile()
        known_models_file_cm = tempfile.NamedTemporaryFile()
        with result_file_cm as result_file, known_models_file_cm as known_models_file:
            if self._determine_django_state_script is not None:
                script = self._determine_django_state_script
            else:
                script = pathlib.Path(
                    str(
                        importlib.resources.files("extended_mypy_django_plugin")
                        / "scripts"
                        / "determine_django_state.py"
                    )
                )

            cmd: list[str] = []

            if script.suffix == ".py":
                cmd.append(sys.executable)
            else:
                with open(script) as fle:
                    line = fle.readline()
                    if line.startswith("#!"):
                        cmd.extend(shlex.split(line[2:]))

            cmd.extend(
                [
                    str(script),
                    "--django-settings-module",
                    self._django_settings_module,
                    "--apps-file",
                    result_file.name,
                    "--known-models-file",
                    known_models_file.name,
                    "--scratch-path",
                    str(scratch_path),
                ]
            )

            try:
                subprocess.run(cmd, capture_output=True, check=True)
            except subprocess.CalledProcessError as err:
                if err.returncode == 2:
                    return 1 if previous_version is None else previous_version
                else:
                    raise

            installed_apps_hash = str(zlib.adler32(pathlib.Path(result_file.name).read_bytes()))
            known_models_hash = str(
                zlib.adler32(pathlib.Path(known_models_file.name).read_bytes())
            )
            return zlib.adler32(
                f"{installed_apps_hash}.{known_models_hash}.{self.lines_hash()}".encode()
            )

    def report_names_getter(
        self,
        installed_apps: list[str],
        model_modules: ModelModules,
        get_model_related_fields: ModelRelatedFieldsGetter,
        get_field_related_model_cls: FieldRelatedModelClsGetter,
    ) -> ReportNamesGetter:
        summaries: dict[str, str] = {}
        installed_apps_hash = f"installed_apps:{zlib.adler32('||'.join(installed_apps).encode())}"
        results, model_children = _DepFinder.find_from(
            model_modules,
            django_settings_module=self._django_settings_module,
            get_model_related_fields=get_model_related_fields,
            get_field_related_model_cls=get_field_related_model_cls,
        )
        for mod, deps in results.items():
            deps_hash = f"deps:{zlib.adler32('||'.join(sorted(deps)).encode())}"
            summaries[mod] = f"{mod} |>> {self._store.prefix}.{installed_apps_hash}.{deps_hash}"

        self._store = self._store.write(summaries)
        self._known_concrete_models.update(model_children)
        return self._get_report_names

    def _get_report_names(self, fullname: str, deps: set[str], /) -> Iterator[str]:
        report = self._store.modules_to_report_name.get(fullname)
        if report is None:
            if fullname.startswith("django.db."):
                return

            if ".models." in fullname or fullname.endswith(".models"):
                report = self._store.add_mod(fullname)

        if report is not None:
            yield report

        if fullname == self._django_settings_module:
            return

        if report or fullname.startswith(f"{self._store.prefix}."):
            yield self._store.modules_to_report_name[self._django_settings_module]
            yield self._django_settings_module

        for dep in deps:
            report = self._store.modules_to_report_name.get(dep)
            if report and report not in deps:
                yield report



from django.http import HttpRequest, HttpResponse, HttpResponseBase

from extended_mypy_django_plugin import Concrete, ConcreteQuerySet, DefaultQuerySet

from .exampleapp.models import Child1, Child2, Parent

T_Child = Concrete.type_var("T_Child", Parent)


def make_child(child: type[T_Child]) -> T_Child:
    return child.objects.create()


def make_any_queryset(child: type[Concrete[Parent]]) -> ConcreteQuerySet[Parent]:
    return child.objects.all()


def make_child1_queryset() -> DefaultQuerySet[Child1]:
    return Child1.objects.all()


def make_child2_queryset() -> DefaultQuerySet[Child2]:
    return Child2.objects.all()


def make_child_typevar_queryset(child: type[T_Child]) -> DefaultQuerySet[T_Child]:
    return child.objects.all()


def ones(model: type[Concrete[Parent]]) -> list[str]:
    # Union[django.db.models.manager.Manager[djangoexample.exampleapp.models.Child1], djangoexample.exampleapp.models.ManagerFromChild2QuerySet[djangoexample.exampleapp.models.Child2], django.db.models.manager.Manager[djangoexample.exampleapp.models.Child3]]
    reveal_type(model.objects)
    return list(model.objects.values_list("one", flat=True))


def index(request: HttpRequest) -> HttpResponseBase:
    made = make_child(Child1)
    # djangoexample.exampleapp.models.Child1
    reveal_type(made)

    any_qs = make_any_queryset(Child1)
    # Union[django.db.models.query._QuerySet[djangoexample.exampleapp.models.Child1], djangoexample.exampleapp.models.Child2QuerySet, django.db.models.query._QuerySet[djangoexample.exampleapp.models.Child3]]
    reveal_type(any_qs)

    qs1 = make_child1_queryset()
    # django.db.models.query._QuerySet[djangoexample.exampleapp.models.Child1]
    reveal_type(qs1)

    qs2 = make_child2_queryset()
    # djangoexample.exampleapp.models.Child2QuerySet
    reveal_type(qs2)
    # djangoexample.exampleapp.models.Child2QuerySet
    reveal_type(qs2.all())
    # djangoexample.exampleapp.models.ManagerFromChild2QuerySet[djangoexample.exampleapp.models.Child2]
    reveal_type(Child2.objects)
    # djangoexample.exampleapp.models.Child2QuerySet[djangoexample.exampleapp.models.Child2]
    reveal_type(Child2.objects.all())

    tvqs1 = make_child_typevar_queryset(Child1)
    # django.db.models.query._QuerySet[djangoexample.exampleapp.models.Child1]
    reveal_type(tvqs1)

    tvqs2 = make_child_typevar_queryset(Child2)
    # djangoexample.exampleapp.models.Child2QuerySet
    reveal_type(tvqs2)

    return HttpResponse("Hello there")



"""
The config class is copied from ``django-stubs`` except it also looks for a
``python_identifier`` option and ensures it's a valid python identifier.
"""

import configparser
import pathlib
import sys
from collections.abc import Callable
from functools import partial
from pathlib import Path
from typing import Any, NoReturn

from mypy_django_plugin.config import (
    COULD_NOT_LOAD_FILE,
    MISSING_DJANGO_SETTINGS,
    MISSING_SECTION,
    DjangoPluginConfig,
    exit_with_error,
)

if sys.version_info[:2] >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

MISSING_SCRATCH_PATH = "missing required 'scratch_path' config"
INVALID_BOOL_SETTING = "invalid {key!r}: the setting must be a boolean"


class Config(DjangoPluginConfig):
    __slots__ = (
        "django_settings_module",
        "strict_settings",
        "scratch_path",
        "determine_django_state_script",
    )
    scratch_path: Path
    determine_django_state_script: Path | None

    def parse_toml_file(self, filepath: pathlib.Path) -> None:
        toml_exit: Callable[[str], NoReturn] = partial(exit_with_error, is_toml=True)
        try:
            with filepath.open(mode="rb") as f:
                data = tomllib.load(f)
        except (tomllib.TOMLDecodeError, OSError):
            toml_exit(COULD_NOT_LOAD_FILE)

        try:
            config: dict[str, Any] = data["tool"]["django-stubs"]
        except KeyError:
            toml_exit(MISSING_SECTION.format(section="tool.django-stubs"))

        if "django_settings_module" not in config:
            toml_exit(MISSING_DJANGO_SETTINGS)

        if "scratch_path" not in config:
            toml_exit(MISSING_SCRATCH_PATH)

        self.django_settings_module = config["django_settings_module"]
        if not isinstance(self.django_settings_module, str):
            toml_exit("invalid 'django_settings_module': the setting must be a string")

        if not isinstance(config["scratch_path"], str):
            toml_exit("invalid 'scratch_path': the setting must be a string")
        self.scratch_path = filepath.parent / config["scratch_path"]
        self.scratch_path.mkdir(parents=True, exist_ok=True)

        if "installed_apps_path" in config:
            if not isinstance(config["determine_django_state_script"], str):
                toml_exit("invalid 'determine_django_state_script': the setting must be a string")
            self.determine_django_state_script = (
                filepath.parent / config["determine_django_state_script"]
            )
        else:
            self.determine_django_state_script = None

        self.strict_settings = config.get("strict_settings", True)
        if not isinstance(self.strict_settings, bool):
            toml_exit(INVALID_BOOL_SETTING.format(key="strict_settings"))

    def parse_ini_file(self, filepath: Path) -> None:
        parser = configparser.ConfigParser()
        try:
            with filepath.open(encoding="utf-8") as f:
                parser.read_file(f, source=str(filepath))
        except OSError:
            exit_with_error(COULD_NOT_LOAD_FILE)

        section = "mypy.plugins.django-stubs"
        if not parser.has_section(section):
            exit_with_error(MISSING_SECTION.format(section=section))

        if not parser.has_option(section, "django_settings_module"):
            exit_with_error(MISSING_DJANGO_SETTINGS)

        if not parser.has_option(section, "scratch_path"):
            exit_with_error(MISSING_SCRATCH_PATH)

        self.django_settings_module = parser.get(section, "django_settings_module").strip("'\"")

        self.scratch_path = filepath.parent / parser.get(section, "scratch_path").strip("'\"")
        self.scratch_path.mkdir(parents=True, exist_ok=True)

        if parser.has_option(section, "determine_django_state_script"):
            self.determine_django_state_script = filepath.parent / parser.get(
                section, "determine_django_state_script"
            ).strip("'\"")
        else:
            self.determine_django_state_script = None

        try:
            self.strict_settings = parser.getboolean(section, "strict_settings", fallback=True)
        except ValueError:
            exit_with_error(INVALID_BOOL_SETTING.format(key="strict_settings"))

    def to_json(self) -> dict[str, Any]:
        """We use this method to reset mypy cache via `report_config_data` hook."""
        return {
            "django_settings_module": self.django_settings_module,
            "scratch_path": str(self.scratch_path),
            "determine_django_state_script": (
                None
                if self.determine_django_state_script is None
                else str(self.determine_django_state_script)
            ),
            "strict_settings": self.strict_settings,
        }



from mypy.nodes import Import, ImportBase, ImportFrom

from ._reports import ModelModules, ReportNamesGetter

Dep = tuple[int, str, int]
DepList = list[tuple[int, str, int]]


class Dependencies:
    """
    This object is used to determine if a model is known, and also what
    dependencies to return for a file for the get_additional_deps hook
    """

    def __init__(
        self, model_modules: ModelModules, report_names_getter: ReportNamesGetter
    ) -> None:
        self._model_modules = model_modules
        self._report_names_getter = report_names_getter

    def is_model_known(self, fullname: str) -> bool:
        for mod, known in self._model_modules.items():
            for cls in known.values():
                if fullname == f"{cls.__module__}.{cls.__qualname__}":
                    return True

        return False

    def for_file(self, fullname: str, imports: list[ImportBase], super_deps: DepList) -> DepList:
        deps = list(super_deps)

        if fullname.startswith("django."):
            return deps

        for imp in imports:
            found: set[str] = set()
            if isinstance(imp, ImportFrom):
                for name in imp.names:
                    found.add(f"{imp.id}.{name[0]}")
            elif isinstance(imp, Import):
                for name in imp.ids:
                    found.add(name[0])

            for full in found:
                for mod in self._model_modules:
                    if full == fullname:
                        continue

                    if full == mod or full.startswith(f"{mod}."):
                        new_dep = (10, mod, -1)
                        if new_dep not in deps:
                            deps.append(new_dep)
                        break

        for report in self._report_names_getter(fullname, {m for _, m, _ in deps}):
            new_dep = (10, report, -1)

            if new_dep not in deps:
                deps.append(new_dep)

        return deps



import enum
import sys
from typing import Generic

from mypy.checker import TypeChecker
from mypy.modulefinder import mypy_path
from mypy.nodes import MypyFile, TypeInfo
from mypy.options import Options
from mypy.plugin import (
    AnalyzeTypeContext,
    AttributeContext,
    DynamicClassDefContext,
    FunctionContext,
)
from mypy.semanal import SemanticAnalyzer
from mypy.typeanal import TypeAnalyser
from mypy.types import CallableType, Instance
from mypy.types import Type as MypyType
from mypy_django_plugin import main
from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.transformers.managers import (
    resolve_manager_method,
    resolve_manager_method_from_instance,
)
from typing_extensions import assert_never

from . import _config, _dependencies, _hook, _reports, _store, actions


class Hook(
    Generic[_hook.T_Ctx, _hook.T_Ret],
    _hook.Hook["ExtendedMypyStubs", _hook.T_Ctx, _hook.T_Ret],
):
    store: _store.Store

    def extra_init(self) -> None:
        self.store = self.plugin.store


class ExtendedMypyStubs(main.NewSemanalDjangoPlugin):
    """
    The ``ExtendedMypyStubs`` mypy plugin extends the
    ``mypy_django_plugin.main.NewSemanalDjangoPlugin`` found in the active python
    environment.

    It implements the following mypy plugin hooks:

    .. automethod:: get_additional_deps

    .. autoattribute:: get_dynamic_class_hook

    .. autoattribute:: get_type_analyze_hook

    .. autoattribute:: get_function_hook

    .. autoattribute:: get_attribute_hook
    """

    plugin_config: _config.Config

    class Annotations(enum.Enum):
        CONCRETE = "extended_mypy_django_plugin.annotations.Concrete"
        CONCRETE_QUERYSET = "extended_mypy_django_plugin.annotations.ConcreteQuerySet"
        DEFAULT_QUERYSET = "extended_mypy_django_plugin.annotations.DefaultQuerySet"

    def __init__(self, options: Options, mypy_version_tuple: tuple[int, int]) -> None:
        super(main.NewSemanalDjangoPlugin, self).__init__(options)
        self.mypy_version_tuple = mypy_version_tuple

        self.plugin_config = _config.Config(options.config_file)
        # Add paths from MYPYPATH env var
        sys.path.extend(mypy_path())
        # Add paths from mypy_path config option
        sys.path.extend(options.mypy_path)

        self.running_in_daemon: bool = "dmypy" in sys.argv[0]
        self.report = _reports.Reports.create(
            determine_django_state_script=self.plugin_config.determine_django_state_script,
            django_settings_module=self.plugin_config.django_settings_module,
            scratch_path=self.plugin_config.scratch_path,
        )

        self.django_context = DjangoContext(self.plugin_config.django_settings_module)
        self.store = _store.Store(
            get_model_class_by_fullname=self.django_context.get_model_class_by_fullname,
            lookup_info=self._lookup_info,
            django_context_model_modules=self.django_context.model_modules,
            is_installed_model=self._is_installed_model,
            known_concrete_models=self.report.known_concrete_models,
        )

        self.dependencies = _dependencies.Dependencies(
            model_modules=self.store.model_modules,
            report_names_getter=self.report.report_names_getter(
                installed_apps=self.django_context.settings.INSTALLED_APPS,
                model_modules=self.store.model_modules,
                get_model_related_fields=self.django_context.get_model_related_fields,
                get_field_related_model_cls=self.django_context.get_field_related_model_cls,
            ),
        )

    def _is_installed_model(self, instance: Instance) -> bool:
        return self.dependencies.is_model_known(instance.type.fullname)

    def _lookup_info(self, fullname: str) -> TypeInfo | None:
        sym = self.lookup_fully_qualified(fullname)
        if sym and isinstance(sym.node, TypeInfo):
            return sym.node
        else:
            return None

    def determine_plugin_version(self, previous_version: int | None = None) -> int:
        """
        Used to set `__version__' where the plugin is defined.

        This lets us tell dmypy to restart itself as necessary.
        """
        if not self.running_in_daemon:
            return 0
        else:
            return self.report.determine_version_hash(
                self.plugin_config.scratch_path, previous_version
            )

    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        """
        Ensure that models are re-analyzed if any other models that depend on
        them change.

        We use a generated "report" to re-analyze a file if a new dependency
        is discovered after this file has been processed.
        """
        results = self.dependencies.for_file(
            file.fullname, imports=file.imports, super_deps=super().get_additional_deps(file)
        )
        return results

    @_hook.hook
    class get_dynamic_class_hook(Hook[DynamicClassDefContext, None]):
        """
        This is used to find ``Concrete.type_var`` and turn that into a ``TypeVar``
        representing each Concrete class of the abstract model provided.

        So say we find::

            T_Child = Concrete.type_var("T_Child", Parent)

        Then we turn that into::

            T_Child = TypeVar("T_Child", Child1, Child2, Child3)
        """

        def choose(self) -> bool:
            class_name, _, method_name = self.fullname.rpartition(".")
            if method_name == "type_var":
                info = self.plugin._get_typeinfo_or_none(class_name)
                if info and info.has_base(ExtendedMypyStubs.Annotations.CONCRETE.value):
                    return True

            return False

        def run(self, ctx: DynamicClassDefContext) -> None:
            assert isinstance(ctx.api, SemanticAnalyzer)

            sem_analyzing = actions.SemAnalyzing(self.store, api=ctx.api)

            return sem_analyzing.transform_type_var_classmethod(ctx)

    @_hook.hook
    class get_type_analyze_hook(Hook[AnalyzeTypeContext, MypyType]):
        """
        Resolve classes annotated with ``Concrete``, ``ConcreteQuerySet`` and
        ``DefaultQuerySet``.
        """

        def choose(self) -> bool:
            return any(
                member.value == self.fullname
                for member in ExtendedMypyStubs.Annotations.__members__.values()
            )

        def run(self, ctx: AnalyzeTypeContext) -> MypyType:
            assert isinstance(ctx.api, TypeAnalyser)
            assert isinstance(ctx.api.api, SemanticAnalyzer)

            Known = ExtendedMypyStubs.Annotations
            name = Known(self.fullname)

            type_analyzer = actions.TypeAnalyzing(self.store, api=ctx.api, sem_api=ctx.api.api)

            if name is Known.CONCRETE:
                method = type_analyzer.find_concrete_models

            elif name is Known.CONCRETE_QUERYSET:
                method = type_analyzer.find_concrete_querysets

            elif name is Known.DEFAULT_QUERYSET:
                method = type_analyzer.find_default_queryset
            else:
                assert_never(name)

            return method(unbound_type=ctx.type)

    @_hook.hook
    class get_function_hook(Hook[FunctionContext, MypyType]):
        """
        Find functions that return a ``DefaultQuerySet`` annotation and resolve
        the annotation.
        """

        def choose(self) -> bool:
            sym = self.plugin.lookup_fully_qualified(self.fullname)
            if not sym or not sym.node:
                return False

            call = getattr(sym.node, "type", None)
            if not isinstance(call, CallableType):
                return False

            return call.is_generic()

        def run(self, ctx: FunctionContext) -> MypyType:
            assert isinstance(ctx.api, TypeChecker)

            type_checking = actions.TypeChecking(self.store, api=ctx.api)

            return type_checking.modify_default_queryset_return_type(
                ctx,
                super_hook=self.super_hook,
                desired_annotation_fullname=ExtendedMypyStubs.Annotations.DEFAULT_QUERYSET.value,
            )

    @_hook.hook
    class get_attribute_hook(Hook[AttributeContext, MypyType]):
        """
        An implementation of the change found in
        https://github.com/typeddjango/django-stubs/pull/2027
        """

        def choose(self) -> bool:
            return self.super_hook is resolve_manager_method

        def run(self, ctx: AttributeContext) -> MypyType:
            assert isinstance(ctx.api, TypeChecker)

            type_checking = actions.TypeChecking(self.store, api=ctx.api)

            return type_checking.extended_get_attribute_resolve_manager_method(
                ctx, resolve_manager_method_from_instance=resolve_manager_method_from_instance
            )



from collections.abc import Callable
from typing import Protocol

from mypy.checker import TypeChecker
from mypy.nodes import CallExpr, MemberExpr, TypeInfo
from mypy.plugin import (
    AttributeContext,
    FunctionContext,
)
from mypy.types import (
    AnyType,
    CallableType,
    FormalArgument,
    Instance,
    TypeOfAny,
    TypeType,
    TypeVarType,
    UnboundType,
    UnionType,
    get_proper_type,
)
from mypy.types import Type as MypyType

from .. import _store


class ResolveManagerMethodFromInstance(Protocol):
    def __call__(
        self, instance: Instance, method_name: str, ctx: AttributeContext
    ) -> MypyType: ...


class TypeChecking:
    def __init__(self, store: _store.Store, *, api: TypeChecker) -> None:
        self.api = api
        self.store = store

    def modify_default_queryset_return_type(
        self,
        ctx: FunctionContext,
        *,
        super_hook: Callable[[FunctionContext], MypyType] | None,
        desired_annotation_fullname: str,
    ) -> MypyType:
        if not isinstance(ctx.default_return_type, UnboundType):
            return ctx.default_return_type

        context = ctx.context
        if not isinstance(context, CallExpr):
            return ctx.default_return_type

        if hasattr(self.api, "get_expression_type"):
            # In later mypy versions
            func = self.api.get_expression_type(context.callee)
        else:
            func = self.api.expr_checker.accept(context.callee)

        func = get_proper_type(func)

        if not isinstance(func, CallableType):
            self.api.fail("Expected to be operating on a callable", context)
            return AnyType(TypeOfAny.from_error)

        if not isinstance(func.ret_type, UnboundType):
            return ctx.default_return_type

        as_generic_type = self.api.named_generic_type(func.ret_type.name, [func.ret_type.args[0]])
        if as_generic_type.type.fullname != desired_annotation_fullname:
            return ctx.default_return_type

        if len(func.ret_type.args) != 1:
            self.api.fail("DefaultQuerySet takes only one argument", context)
            return AnyType(TypeOfAny.from_error)

        found_type: MypyType | None = None

        type_var = func.ret_type.args[0]

        if isinstance(type_var, UnboundType):
            match: FormalArgument | None = None
            for arg in func.formal_arguments():
                arg_name: str | None = None
                formal_arg_type = get_proper_type(arg.typ)
                if isinstance(formal_arg_type, TypeType) and isinstance(
                    formal_arg_type.item, TypeVarType
                ):
                    arg_name = formal_arg_type.item.name

                elif isinstance(arg, TypeVarType):
                    arg_name = formal_arg_type.name

                if arg_name and arg_name == type_var.name:
                    match = arg
                    break

            if match is not None:
                for arg_name, arg_type in zip(ctx.callee_arg_names, ctx.arg_types):
                    if arg_name == match.name:
                        found_type = arg_type[0]

            if found_type is None:
                self.api.fail("Failed to find an argument that matched the type var", context)
                return AnyType(TypeOfAny.from_error)

            found_type = get_proper_type(found_type)
            if isinstance(found_type, CallableType):
                type_var = found_type.ret_type
            else:
                type_var = found_type

        if not isinstance(type_var, Instance | UnionType):
            self.api.fail("Don't know what to do with what DefaultQuerySet was given", context)
            return AnyType(TypeOfAny.from_error)

        try:
            querysets = tuple(self.store.realise_querysets(type_var, self.lookup_info))
        except _store.RestartDmypy:
            self.api.fail("You probably need to restart dmypy", context)
            return AnyType(TypeOfAny.from_error)
        except _store.UnionMustBeOfTypes:
            self.api.fail("Union must be of instances of models", context)
            return AnyType(TypeOfAny.from_error)
        else:
            return UnionType(querysets)

    def extended_get_attribute_resolve_manager_method(
        self,
        ctx: AttributeContext,
        *,
        resolve_manager_method_from_instance: ResolveManagerMethodFromInstance,
    ) -> MypyType:
        """
        Copied from django-stubs after https://github.com/typeddjango/django-stubs/pull/2027

        A 'get_attribute_hook' that is intended to be invoked whenever the TypeChecker encounters
        an attribute on a class that has 'django.db.models.BaseManager' as a base.
        """
        # Skip (method) type that is currently something other than Any of type `implementation_artifact`
        default_attr_type = get_proper_type(ctx.default_attr_type)
        if not isinstance(default_attr_type, AnyType):
            return default_attr_type
        elif default_attr_type.type_of_any != TypeOfAny.implementation_artifact:
            return default_attr_type

        # (Current state is:) We wouldn't end up here when looking up a method from a custom _manager_.
        # That's why we only attempt to lookup the method for either a dynamically added or reverse manager.
        if isinstance(ctx.context, MemberExpr):
            method_name = ctx.context.name
        elif isinstance(ctx.context, CallExpr) and isinstance(ctx.context.callee, MemberExpr):
            method_name = ctx.context.callee.name
        else:
            ctx.api.fail("Unable to resolve return type of queryset/manager method", ctx.context)
            return AnyType(TypeOfAny.from_error)

        if isinstance(ctx.type, Instance):
            return resolve_manager_method_from_instance(
                instance=ctx.type, method_name=method_name, ctx=ctx
            )
        elif isinstance(ctx.type, UnionType) and all(
            isinstance(get_proper_type(instance), Instance) for instance in ctx.type.items
        ):
            items: list[Instance] = []
            for instance in ctx.type.items:
                inst = get_proper_type(instance)
                if isinstance(inst, Instance):
                    items.append(inst)

            resolved = tuple(
                resolve_manager_method_from_instance(
                    instance=inst, method_name=method_name, ctx=ctx
                )
                for inst in items
            )
            return UnionType(resolved)
        else:
            ctx.api.fail(
                f'Unable to resolve return type of queryset/manager method "{method_name}"',
                ctx.context,
            )
            return AnyType(TypeOfAny.from_error)

    def lookup_info(self, fullname: str) -> TypeInfo | None:
        return self.store._plugin_lookup_info(fullname)



from __future__ import annotations

from collections.abc import Sequence
from typing import Generic, TypeVar

from django.db import models

T_Parent = TypeVar("T_Parent", bound=models.Model)


class Concrete(Generic[T_Parent]):
    """
    The ``Concrete`` annotation exists as a class with functionality for both
    runtime and static type checking time.

    At runtime it can be used to create special ``TypeVar`` objects that may
    represent any one of the concrete children of some abstract class and
    it can be used to find those concrete children.

    At static type checking time (specifically with ``mypy``) it is used to create
    a type that represents the Union of all the concrete children of some
    abstract model.

    .. automethod:: find_children
    .. automethod:: type_var
    """

    @classmethod
    def find_children(cls, parent: type[models.Model]) -> Sequence[type[models.Model]]:
        """
        At runtime this will find all the concrete children of some model.

        That is all models that inherit from this model and aren't abstract
        themselves.
        """
        found: list[type[models.Model]] = []

        from django.contrib.contenttypes.models import ContentType

        content_types = ContentType.objects.filter(app_label=parent._meta.app_label)
        for ct in content_types:
            model = ct.model_class()
            if model is None:
                continue
            if not issubclass(model, parent):
                continue
            if hasattr(model, "Meta") and getattr(model.Meta, "is_abstract", False):
                continue
            found.append(model)

        return found

    @classmethod
    def type_var(cls, name: str, parent: type[models.Model]) -> TypeVar:
        """
        This returns an empty ``TypeVar`` at runtime, but the ``mypy`` plugin will
        recognise that this ``TypeVar`` represents a choice of all the concrete
        children of the specified model.
        """
        return TypeVar(name)


class ConcreteQuerySet(Generic[T_Parent]):
    """
    This is used to annotate a model such that the mypy plugin may turn this into
    a union of all the default querysets for all the concrete children of the
    specified abstract model class.
    """


class DefaultQuerySet(Generic[T_Parent]):
    """
    This is used to annotate a model such that the mypy plugin may turn this into
    the queryset type used by the default manager on the model.
    """



from django.db import models

from djangoexample.exampleapp.models import Parent


class ChildOther(Parent):
    two = models.CharField(max_length=60)


class ChildOther2(Parent):
    two = models.CharField(max_length=60)



from collections.abc import Iterator, Mapping, Sequence
from typing import Protocol

from django.db import models
from mypy.nodes import TypeInfo
from mypy.types import Instance, UnionType, get_proper_type
from mypy.types import Type as MypyType

from ._reports import ModelModules

QUERYSET_CLASS_FULLNAME = "django.db.models.query._QuerySet"
MODEL_CLASS_FULLNAME = "django.db.models.base.Model"


class UnionMustBeOfTypes(Exception):
    pass


class RestartDmypy(Exception):
    pass


class LookupFunction(Protocol):
    def __call__(self, fullname: str) -> TypeInfo | None: ...


class LookupInstanceFunction(Protocol):
    def __call__(self, fullname: str) -> Instance | None: ...


class GetModelClassByFullname(Protocol):
    def __call__(self, fullname: str) -> type[models.Model] | None: ...


class IsInstalledModel(Protocol):
    def __call__(self, instance: Instance) -> bool: ...


class KnownConcreteModelsGetter(Protocol):
    def __call__(self, fullname: str) -> set[str]: ...


class Store:
    """
    The store is used to interrogate the metadata on ``TypeInfo`` objects to determine
    the available concrete models and queryset objects when resolving the annotations
    this plugin provides.

    .. automethod:: retrieve_concrete_children_types

    .. automethod:: realise_querysets
    """

    def __init__(
        self,
        get_model_class_by_fullname: GetModelClassByFullname,
        lookup_info: LookupFunction,
        django_context_model_modules: Mapping[str, object],
        is_installed_model: IsInstalledModel,
        known_concrete_models: KnownConcreteModelsGetter,
    ) -> None:
        self._get_model_class_by_fullname = get_model_class_by_fullname
        self._plugin_lookup_info = lookup_info
        self._django_context_model_modules = django_context_model_modules
        self._is_installed_model = is_installed_model
        self._known_concrete_models = known_concrete_models
        self.model_modules = self._determine_model_modules()

    def retrieve_concrete_children_types(
        self,
        parent: TypeInfo,
        lookup_info: LookupFunction,
        lookup_instance: LookupInstanceFunction,
    ) -> Sequence[MypyType]:
        """
        Given a ``TypeInfo`` representing some model, return ``MypyType`` objects
        for all the concrete children related to the specified model.
        """
        values: list[MypyType] = []

        concrete_type_infos = self._retrieve_concrete_children_info_from_metadata(
            parent, lookup_info
        )
        for info in concrete_type_infos:
            instance = lookup_instance(info.fullname)
            if instance and self._is_installed_model(instance):
                values.append(instance)

        return values

    def realise_querysets(
        self, type_var: Instance | UnionType, lookup_info: LookupFunction
    ) -> Iterator[Instance]:
        """
        Given either a specific model, or a union of models, return the
        default querysets for those models.
        """
        querysets = self._get_queryset_fullnames(type_var, lookup_info)
        for fullname, model in querysets:
            queryset = lookup_info(fullname)
            if not queryset:
                raise RestartDmypy()

            if not queryset.is_generic():
                yield Instance(queryset, [])
            else:
                yield Instance(
                    queryset,
                    [Instance(model, []) for _ in range(len(queryset.type_vars))],
                )

    def _determine_model_modules(self) -> ModelModules:
        """
        Old version of django-stubs has this as a different datastructure
        """
        result: dict[str, dict[str, type[models.Model]]] = {}
        for k, v in self._django_context_model_modules.items():
            if isinstance(v, dict):
                result[k] = v
            elif isinstance(v, set):
                result[k] = {
                    cls.__name__: cls
                    for cls in v
                    if isinstance(cls, type) and issubclass(cls, models.Model)
                }
        return result

    def _retrieve_concrete_children_info_from_metadata(
        self, parent: TypeInfo, lookup_info: LookupFunction
    ) -> Sequence[TypeInfo]:
        """
        For the children recorded in the metadata for this model, return those
        that aren't abstract
        """
        children = sorted(self._known_concrete_models(parent.fullname))

        ret: list[TypeInfo] = []
        for child in children:
            info = lookup_info(child)
            if not info:
                continue

            abstract: bool = False
            if "django" not in info.metadata:
                # Old versions of mypy/django-stubs don't have metadata at this point
                model_cls = self._get_model_class_by_fullname(info.fullname)
                abstract = bool(model_cls and model_cls._meta.abstract)
            else:
                abstract = info.metadata.get("django", {}).get("is_abstract_model", False)

            if not abstract:
                ret.append(info)

        return ret

    def _get_queryset_fullnames(
        self, type_var: Instance | UnionType, lookup_info: LookupFunction
    ) -> Iterator[tuple[str, TypeInfo]]:
        """
        Return the fullnames of the default querysets for the models represented
        by this instance or Union of instances.
        """
        children: list[TypeInfo] = []
        if isinstance(type_var, UnionType):
            for item in type_var.items:
                item = get_proper_type(item)
                if not isinstance(item, Instance):
                    raise UnionMustBeOfTypes()
                children.append(item.type)
        else:
            children.append(type_var.type)

        for child in children:
            yield (
                self._get_dynamic_queryset_fullname(child, lookup_info) or QUERYSET_CLASS_FULLNAME,
                child,
            )

    def _get_dynamic_manager(
        self, model: TypeInfo, lookup_info: LookupFunction
    ) -> TypeInfo | None:
        """
        For some model return a custom manager if one exists
        """
        model_cls = self._get_model_class_by_fullname(model.fullname)
        if model_cls is None:
            raise RestartDmypy()

        manager = model_cls._default_manager
        if manager is None:
            return None

        if not isinstance(manager, models.Manager):
            return None

        manager_fullname = manager.__class__.__module__ + "." + manager.__class__.__qualname__
        manager_info = lookup_info(manager_fullname)
        if manager_info is None:
            base_manager_class = manager.__class__.__bases__[0]
            base_manager_fullname = (
                base_manager_class.__module__ + "." + base_manager_class.__qualname__
            )

            base_manager_info = lookup_info(base_manager_fullname)
            if not base_manager_info:
                raise RestartDmypy()

            metadata = base_manager_info.metadata

            generated_managers: dict[str, str]
            if "from_queryset_managers" not in metadata:
                metadata["from_queryset_managers"] = {}
            generated_managers = {
                k: v for k, v in metadata["from_queryset_managers"].items() if isinstance(v, str)
            }

            generated_manager_name: str | None = generated_managers.get(manager_fullname)
            if generated_manager_name is None:
                return None

            manager_info = lookup_info(generated_manager_name)

        return manager_info

    def _get_dynamic_queryset_fullname(
        self, model: TypeInfo, lookup_info: LookupFunction
    ) -> str | None:
        """
        For this model, return the fullname of the custom queryset for the
        default manager if there is such a custom QuerySet.
        """
        dynamic_manager = self._get_dynamic_manager(model, lookup_info)
        if not dynamic_manager:
            model_cls = self._get_model_class_by_fullname(model.fullname)
            if (
                model_cls
                and hasattr(model_cls, "_default_manager")
                and isinstance(model_cls._default_manager, models.Manager)
                and hasattr(model_cls._default_manager, "_queryset_class")
            ):
                queryset = model_cls._default_manager._queryset_class
                if isinstance(queryset, type) and issubclass(queryset, models.QuerySet):
                    return queryset.__module__ + "." + queryset.__qualname__
            return None

        name = dynamic_manager.metadata["django"].get("from_queryset_manager")
        if name is not None and isinstance(name, str):
            assert isinstance(name, str)
            return name

        return None



from mypy.nodes import (
    GDEF,
    StrExpr,
    SymbolTableNode,
    TypeInfo,
    TypeVarExpr,
)
from mypy.plugin import (
    DynamicClassDefContext,
)
from mypy.semanal import SemanticAnalyzer
from mypy.types import (
    AnyType,
    TypeOfAny,
    TypeVarType,
)

from .. import _store


class SemAnalyzing:
    def __init__(self, store: _store.Store, *, api: SemanticAnalyzer) -> None:
        self.api = api
        self.store = store

    def transform_type_var_classmethod(self, ctx: DynamicClassDefContext) -> None:
        if not isinstance(ctx.call.args[0], StrExpr):
            self.api.fail(
                "First argument to Concrete.type_var must be a string of the name of the variable",
                ctx.call,
            )
            return

        name = ctx.call.args[0].value
        if name != ctx.name:
            self.api.fail(
                f"First argument {name} was not the name of the variable {ctx.name}",
                ctx.call,
            )
            return

        module = self.api.modules[self.api.cur_mod_id]
        if isinstance(module.names.get(name), TypeVarType):
            return

        parent: SymbolTableNode | None = None
        try:
            parent = self.api.lookup_type_node(ctx.call.args[1])
        except AssertionError:
            parent = None

        if parent is None:
            self.api.fail(
                "Second argument to Concrete.type_var must be the abstract model class to find concrete instances of",
                ctx.call,
            )
            return

        if not isinstance(parent.node, TypeInfo):
            self.api.fail(
                "Second argument to Concrete.type_var was not pointing at a class", ctx.call
            )
            return

        object_type = self.api.named_type("builtins.object")
        values = self.store.retrieve_concrete_children_types(
            parent.node, self.lookup_info, self.api.named_type_or_none
        )
        if not values:
            self.api.fail(f"No concrete children found for {parent.node.fullname}", ctx.call)

        type_var_expr = TypeVarExpr(
            name=name,
            fullname=f"{self.api.cur_mod_id}.{name}",
            values=list(values),
            upper_bound=object_type,
            default=AnyType(TypeOfAny.from_omitted_generics),
        )

        module.names[name] = SymbolTableNode(GDEF, type_var_expr, plugin_generated=True)
        return None

    def lookup_info(self, fullname: str) -> TypeInfo | None:
        instance = self.api.named_type_or_none(fullname)
        if instance:
            return instance.type

        return self.store._plugin_lookup_info(fullname)
